"""This sub-package contains I/O-related modules such as specific format backends."""

# All I/O modules are lazy-loaded - they're imported when accessed
# This avoids circular import issues and improves performance
